%{
#include <cstdio>
#include <cstdlib>
#include <cmath>
#include <string>
#include <utility>
#include "scanner_shim.hpp"
#include "quant_c_parse.hpp"
#include "util/pos.hpp"
#include "util/type_def_set.hpp"
#include "token/token.hpp"

#define YY_DECL \
    int quant_c_compiler::Scanner::lex( \
        quant_c_compiler::Parser::semantic_type *yylval \
        )

using namespace util;
using namespace quant_c_token;
using namespace quant_c_compiler;
#define TOKEN Parser::token_type

/* Function delcarations */
void count(const char* yytext);
bool ends_with(std::string const &string, std::string const &phrase);
void decode_int(const char * yytext, Parser::semantic_type *yylval, int base);
void decode_float(const char * yytext, Parser::semantic_type *yylval);

%}

%option c++ interactive noyywrap noyylineno nodefault

DIGIT        [0-9]
LETTER       [a-zA-Z_]
HEX          [a-fA-F0-9]
EXP          [Ee][+-]?{DIGIT}+
FLOAT_SIZE   (f|F|l|L)
INT_SIZE     (u|U|l|L)*

ID          {LETTER}({LETTER}|{DIGIT})*
CTRL        "+C"
INV         "~>"



%%
    /* Comments */
"//"[^\n]*$            {count(YYText());}
"/*"([^*]|(\*+[^*/]))*\*+\/  {count(YYText());}

    /* Reserved words */
"auto"          { count(YYText());
                  yylval->emplace<std::pair<int, int> >
                  (std::make_pair(Pos::linenum, Pos::colnum));
                  return(TOKEN::AUTO);
                }
"break"         { count(YYText());
                  yylval->emplace<std::pair<int, int> >
                    (std::make_pair(Pos::linenum, Pos::colnum));
                  return(TOKEN::BREAK);
                }
"case"          { count(YYText());
                  yylval->emplace<std::pair<int, int> >
                    (std::make_pair(Pos::linenum, Pos::colnum));
                  return(TOKEN::CASE);
                }
"char"          { count(YYText());
                  yylval->emplace<std::pair<int, int> >
                    (std::make_pair(Pos::linenum, Pos::colnum));
                  return(TOKEN::CHAR);
                }
"const"         { count(YYText());
                  yylval->emplace<std::pair<int, int> >
                    (std::make_pair(Pos::linenum, Pos::colnum));
                  return(TOKEN::CONST);
                }
"continue"      { count(YYText());
                  yylval->emplace<std::pair<int, int> >
                    (std::make_pair(Pos::linenum, Pos::colnum));
                  return(TOKEN::CONTINUE);
                }
"default"       { count(YYText());
                  yylval->emplace<std::pair<int, int> >
                    (std::make_pair(Pos::linenum, Pos::colnum));
                  return(TOKEN::DEFAULT);
                }
"do"            { count(YYText());
                  yylval->emplace<std::pair<int, int> >
                    (std::make_pair(Pos::linenum, Pos::colnum));
                  return(TOKEN::DO);
                }
"double"        { count(YYText());
                  yylval->emplace<std::pair<int, int> >
                    (std::make_pair(Pos::linenum, Pos::colnum));
                  return(TOKEN::DOUBLE);
                }
"else"          { count(YYText());
                  yylval->emplace<std::pair<int, int> >
                    (std::make_pair(Pos::linenum, Pos::colnum));
                  return(TOKEN::ELSE);
                }
"enum"          { count(YYText());
                  yylval->emplace<std::pair<int, int> >
                    (std::make_pair(Pos::linenum, Pos::colnum));
                  return(TOKEN::ENUM);
                }
"extern"        { count(YYText());
                  yylval->emplace<std::pair<int, int> >
                    (std::make_pair(Pos::linenum, Pos::colnum));
                  return(TOKEN::EXTERN);
                }
"float"         { count(YYText());
                  yylval->emplace<std::pair<int, int> >
                    (std::make_pair(Pos::linenum, Pos::colnum));
                  return(TOKEN::FLOAT);
                }
"for"           { count(YYText());
                  yylval->emplace<std::pair<int, int> >
                    (std::make_pair(Pos::linenum, Pos::colnum));
                  return(TOKEN::FOR);
                }
"goto"          { count(YYText());
                  yylval->emplace<std::pair<int, int> >
                    (std::make_pair(Pos::linenum, Pos::colnum));
                  return(TOKEN::GOTO);
                }
"if"            { count(YYText());
                  yylval->emplace<std::pair<int, int> >
                    (std::make_pair(Pos::linenum, Pos::colnum));
                  return(TOKEN::IF);
                }
"int"           { count(YYText());
                  yylval->emplace<std::pair<int, int> >
                    (std::make_pair(Pos::linenum, Pos::colnum));
                  return(TOKEN::INT);
                }
"long"          { count(YYText());
                  yylval->emplace<std::pair<int, int> >
                    (std::make_pair(Pos::linenum, Pos::colnum));
                  return(TOKEN::LONG);
                }
"register"      { count(YYText());
                  yylval->emplace<std::pair<int, int> >
                    (std::make_pair(Pos::linenum, Pos::colnum));
                  return(TOKEN::REGISTER);
                }
"return"        { count(YYText());
                  yylval->emplace<std::pair<int, int> >
                    (std::make_pair(Pos::linenum, Pos::colnum));
                  return(TOKEN::RETURN);
                }
"short"         { count(YYText());
                  yylval->emplace<std::pair<int, int> >
                    (std::make_pair(Pos::linenum, Pos::colnum));
                  return(TOKEN::SHORT);
                }
"signed"        { count(YYText());
                  yylval->emplace<std::pair<int, int> >
                    (std::make_pair(Pos::linenum, Pos::colnum));
                  return(TOKEN::SIGNED);
                }
"sizeof"        { count(YYText());
                  yylval->emplace<std::pair<int, int> >
                    (std::make_pair(Pos::linenum, Pos::colnum));
                  return(TOKEN::SIZEOF);
                }
"static"        { count(YYText());
                  yylval->emplace<std::pair<int, int> >
                    (std::make_pair(Pos::linenum, Pos::colnum));
                  return(TOKEN::STATIC);
                }
"struct"        { count(YYText());
                  yylval->emplace<std::pair<int, int> >
                    (std::make_pair(Pos::linenum, Pos::colnum));
                  return(TOKEN::STRUCT);
                }
"switch"        { count(YYText());
                  yylval->emplace<std::pair<int, int> >
                    (std::make_pair(Pos::linenum, Pos::colnum));
                  return(TOKEN::SWITCH);
                }
"typedef"       { count(YYText());
                  yylval->emplace<std::pair<int, int> >
                    (std::make_pair(Pos::linenum, Pos::colnum));
                  return(TOKEN::TYPEDEF);
                }
"union"         { count(YYText());
                  yylval->emplace<std::pair<int, int> >
                    (std::make_pair(Pos::linenum, Pos::colnum));
                  return(TOKEN::UNION);
                }
"unsigned"      { count(YYText());
                  yylval->emplace<std::pair<int, int> >
                    (std::make_pair(Pos::linenum, Pos::colnum));
                  return(TOKEN::UNSIGNED);
                }
"void"          { count(YYText());
                  yylval->emplace<std::pair<int, int> >
                    (std::make_pair(Pos::linenum, Pos::colnum));
                  return(TOKEN::VOID);
                }
"volatile"      { count(YYText());
                  yylval->emplace<std::pair<int, int> >
                    (std::make_pair(Pos::linenum, Pos::colnum));
                  return(TOKEN::VOLATILE);
                }
"while"         { count(YYText());
                  yylval->emplace<std::pair<int, int> >
                    (std::make_pair(Pos::linenum, Pos::colnum));
                  return(TOKEN::WHILE);
                }
"quantum"       { count(YYText());
                  yylval->emplace<std::pair<int, int> >
                    (std::make_pair(Pos::linenum, Pos::colnum));
                  return(TOKEN::QUANTUM);
                }
"bool"          { count(YYText());
                  yylval->emplace<std::pair<int, int> >
                    (std::make_pair(Pos::linenum, Pos::colnum));
                  return(TOKEN::BOOL);
                }

    /* Integer literals */
0[xX]{HEX}+{INT_SIZE}{0,3}      {
                                    int base = 16;
                                    count(YYText());
                                    decode_int(YYText(), yylval, base);
                                    return(TOKEN::INT_LITERAL);
                                }
0{DIGIT}+{INT_SIZE}{0,3}        {
                                    int base = 8;
                                    count(YYText());
                                    decode_int(YYText(), yylval, base);
                                    return(TOKEN::INT_LITERAL);
                                }
{DIGIT}+{INT_SIZE}{0,3}         {
                                    int base = 10;
                                    count(YYText());
                                    decode_int(YYText(), yylval, base);
                                    return(TOKEN::INT_LITERAL);
                                }

    /* Character literals */
'\\.'               {
                                count(YYText());
                                std::string char_str = YYText();

                                // Determine escape-char value based on the
                                // character after the forward slash
                                char value;
                                switch (char_str[2]) {
                                    case 'a':  value = '\a'; break;
                                    case 'b':  value = '\b'; break;
                                    case 'f':  value = '\f'; break;
                                    case 'n':  value = '\n'; break;
                                    case 'r':  value = '\r'; break;
                                    case 't':  value = '\t'; break;
                                    case 'v':  value = '\v'; break;
                                    case '\\': value = '\\'; break;
                                    case '\'':  value = '\''; break;
                                    case '"': value = '"'; break;
                                    case '?':  value = '\?'; break;
                                    default:
                                        throw InvalidTokenException(char_str, Pos::linenum, Pos::colnum);
                                }
                                yylval->emplace<CharLitToken>(
                                    CharLitToken(value,
                                        Pos::linenum, Pos::colnum));
                                return(TOKEN::CHAR_LITERAL);
                              }
'[0x20-0x7E]'               {
                              count(YYText());
                              std::string char_str = YYText();
                              char value = char_str[1];
                              yylval->emplace<CharLitToken>(
                                  CharLitToken(value,
                                      Pos::linenum, Pos::colnum));
                              return(TOKEN::CHAR_LITERAL);
                            }

    /* Floating point literals */
{DIGIT}+{EXP}{FLOAT_SIZE}                   {
                                                count(YYText());
                                                decode_float(YYText(), yylval);
                                                return(TOKEN::FLOAT_LITERAL);
                                            }

{DIGIT}*"."{DIGIT}+({EXP})?{FLOAT_SIZE}?    {
                                                count(YYText());
                                                decode_float(YYText(), yylval);
                                                return(TOKEN::FLOAT_LITERAL);
                                            }

{DIGIT}+"."{DIGIT}*({EXP})?{FLOAT_SIZE}?    {
                                                count(YYText());
                                                decode_float(YYText(), yylval);
                                                return(TOKEN::FLOAT_LITERAL);
                                            }

"PI"                        {
                                count(YYText());
                                yylval->emplace<FloatLitToken>(
                                FloatLitToken(M_PI,
                                              FloatLitSize::DOUBLE,
                                              Pos::linenum,
                                              Pos::colnum
                                             )
                                );
                                return(TOKEN::FLOAT_LITERAL);
                            }
    /* Boolean Literals */
(true|false)               {
                            count(YYText());
                            std::string yytext_str = YYText();
                            yylval->emplace<BoolLitToken>(
                                BoolLitToken(yytext_str == "true",
                                             Pos::linenum,
                                             Pos::colnum
                                             )
                            );
                            return(TOKEN::BOOL_LITERAL);
                           }

    /* String Literals */
\"(\\.|[^\\"])*\"          {
                                count(YYText());
                                std::string yytext_str = YYText();
                                yylval->emplace<StringLitToken>(
                                    YYText(),
                                    Pos::linenum,
                                    Pos::colnum
                                );
                                return(TOKEN::STRING_LITERAL);
                           }

    /* Identifiers */
{ID}            {
                   std::string name = YYText();
                   count(name.c_str());
                   if (LexFeedback::type_def_set.find(name) ==
                       LexFeedback::type_def_set.end()) {
                    yylval->emplace<IdentifierToken>(
                                          IdentifierToken(YYText(),
                                          Pos::linenum, Pos::colnum
                                      ));
                    return(TOKEN::IDENTIFIER);
                   } else {
                     yylval->emplace<TypeNameToken>(
                                          TypeNameToken(YYText(),
                                          Pos::linenum, Pos::colnum
                                      ));
                     return(TOKEN::TYPE_NAME);
                   }


               }

    /* Quantum Operations */
"%I"           { count(YYText());
                 yylval->emplace<std::pair<int, int> >
                    (std::make_pair(Pos::linenum, Pos::colnum));
                 return(TOKEN::IDENTITY);
               }
"%X"           { count(YYText());
                yylval->emplace<std::pair<int, int> >
                   (std::make_pair(Pos::linenum, Pos::colnum));
                return(TOKEN::X_GATE);
               }
"%Y"           { count(YYText());
                 yylval->emplace<std::pair<int, int> >
                    (std::make_pair(Pos::linenum, Pos::colnum));
                 return(TOKEN::Y_GATE);
               }
"%Z"           { count(YYText());
                 yylval->emplace<std::pair<int, int> >
                    (std::make_pair(Pos::linenum, Pos::colnum));
                 return(TOKEN::Z_GATE);
               }
"%H"           { count(YYText());
                 yylval->emplace<std::pair<int, int> >
                    (std::make_pair(Pos::linenum, Pos::colnum));
                 return(TOKEN::HADAMARD);
               }
"%PHASE"       { count(YYText());
                 yylval->emplace<std::pair<int, int> >
                    (std::make_pair(Pos::linenum, Pos::colnum));
                 return(TOKEN::PHASE);
               }
"%S"           { count(YYText());
                 yylval->emplace<std::pair<int, int> >
                    (std::make_pair(Pos::linenum, Pos::colnum));
                 return(TOKEN::S_GATE);
               }
"%T"           { count(YYText());
                 yylval->emplace<std::pair<int, int> >
                    (std::make_pair(Pos::linenum, Pos::colnum));
                 return(TOKEN::T_GATE);
               }
"%RX"          { count(YYText());
                 yylval->emplace<std::pair<int, int> >
                    (std::make_pair(Pos::linenum, Pos::colnum));
                 return(TOKEN::RX);
               }
"%RY"          { count(YYText());
                 yylval->emplace<std::pair<int, int> >
                    (std::make_pair(Pos::linenum, Pos::colnum));
                 return(TOKEN::RY);
               }
"%RZ"          { count(YYText());
                 yylval->emplace<std::pair<int, int> >
                    (std::make_pair(Pos::linenum, Pos::colnum));
                 return(TOKEN::RZ);
               }
"<>"          { count(YYText());
                 yylval->emplace<std::pair<int, int> >
                    (std::make_pair(Pos::linenum, Pos::colnum));
                 return(TOKEN::SWAP);
               }
"<P>"          { count(YYText());
                 yylval->emplace<std::pair<int, int> >
                    (std::make_pair(Pos::linenum, Pos::colnum));
                 return(TOKEN::PSWAP);
               }
"<I>"          { count(YYText());
                 yylval->emplace<std::pair<int, int> >
                    (std::make_pair(Pos::linenum, Pos::colnum));
                 return(TOKEN::ISWAP);
               }
"%M"           { count(YYText());
                 yylval->emplace<std::pair<int, int> >
                    (std::make_pair(Pos::linenum, Pos::colnum));
                 return(TOKEN::MEASURE);
               }
{CTRL}         { count(YYText());
                 yylval->emplace<std::pair<int, int> >
                    (std::make_pair(Pos::linenum, Pos::colnum));
                 return(TOKEN::CONTROL);
               }
{INV}          { count(YYText());
                 yylval->emplace<std::pair<int, int> >
                    (std::make_pair(Pos::linenum, Pos::colnum));
                 return(TOKEN::INVERSE);
               }

    /* Standard C Symbols */
"..."          { count(YYText());
                 yylval->emplace<std::pair<int, int> >
                 (std::make_pair(Pos::linenum, Pos::colnum));
                 return(TOKEN::ELLIPSIS);
               }
">>="          { count(YYText());
                 yylval->emplace<std::pair<int, int> >
                 (std::make_pair(Pos::linenum, Pos::colnum));
                 return(TOKEN::RIGHT_ASSIGN);
               }
"<<="          { count(YYText());
                 yylval->emplace<std::pair<int, int> >
                 (std::make_pair(Pos::linenum, Pos::colnum));
                 return(TOKEN::LEFT_ASSIGN);
               }
"+="           { count(YYText());
                 yylval->emplace<std::pair<int, int> >
                 (std::make_pair(Pos::linenum, Pos::colnum));
                 return(TOKEN::ADD_ASSIGN);
               }
"-="           { count(YYText());
                 yylval->emplace<std::pair<int, int> >
                 (std::make_pair(Pos::linenum, Pos::colnum));
                 return(TOKEN::SUB_ASSIGN);
               }
"*="           { count(YYText());
                 yylval->emplace<std::pair<int, int> >
                 (std::make_pair(Pos::linenum, Pos::colnum));
                 return(TOKEN::MUL_ASSIGN);
               }
"/="           { count(YYText());
                 yylval->emplace<std::pair<int, int> >
                 (std::make_pair(Pos::linenum, Pos::colnum));
                 return(TOKEN::DIV_ASSIGN);
               }
"%="           { count(YYText());
                 yylval->emplace<std::pair<int, int> >
                 (std::make_pair(Pos::linenum, Pos::colnum));
                 return(TOKEN::MOD_ASSIGN);
               }
"&="           { count(YYText());
                 yylval->emplace<std::pair<int, int> >
                 (std::make_pair(Pos::linenum, Pos::colnum));
                 return(TOKEN::AND_ASSIGN);
               }
"^="           { count(YYText());
                 yylval->emplace<std::pair<int, int> >
                 (std::make_pair(Pos::linenum, Pos::colnum));
                 return(TOKEN::XOR_ASSIGN);
               }
"|="           { count(YYText());
                 yylval->emplace<std::pair<int, int> >
                 (std::make_pair(Pos::linenum, Pos::colnum));
                 return(TOKEN::OR_ASSIGN);
               }
">>"           { count(YYText());
                 yylval->emplace<std::pair<int, int> >
                 (std::make_pair(Pos::linenum, Pos::colnum));
                 return(TOKEN::RIGHT_OP);
               }
"<<"           { count(YYText());
                 yylval->emplace<std::pair<int, int> >
                 (std::make_pair(Pos::linenum, Pos::colnum));
                 return(TOKEN::LEFT_OP);
               }
"++"           { count(YYText());
                 yylval->emplace<std::pair<int, int> >
                 (std::make_pair(Pos::linenum, Pos::colnum));
                 return(TOKEN::INC_OP);
               }
"--"           { count(YYText());
                 yylval->emplace<std::pair<int, int> >
                 (std::make_pair(Pos::linenum, Pos::colnum));
                 return(TOKEN::DEC_OP);
               }
"->"           { count(YYText());
                 yylval->emplace<std::pair<int, int> >
                 (std::make_pair(Pos::linenum, Pos::colnum));
                 return(TOKEN::PTR_OP);
               }
"&&"           { count(YYText());
                 yylval->emplace<std::pair<int, int> >
                 (std::make_pair(Pos::linenum, Pos::colnum));
                 return(TOKEN::AND_OP);
               }
"||"           { count(YYText());
                 yylval->emplace<std::pair<int, int> >
                 (std::make_pair(Pos::linenum, Pos::colnum));
                 return(TOKEN::OR_OP);
               }
"<="           { count(YYText());
                 yylval->emplace<std::pair<int, int> >
                 (std::make_pair(Pos::linenum, Pos::colnum));
                 return(TOKEN::LE_OP);
               }
">="           { count(YYText());
                 yylval->emplace<std::pair<int, int> >
                 (std::make_pair(Pos::linenum, Pos::colnum));
                 return(TOKEN::GE_OP);
               }
"=="           { count(YYText());
                 yylval->emplace<std::pair<int, int> >
                 (std::make_pair(Pos::linenum, Pos::colnum));
                 return(TOKEN::EQ_OP);
               }
"!="           { count(YYText());
                 yylval->emplace<std::pair<int, int> >
                 (std::make_pair(Pos::linenum, Pos::colnum));
                 return(TOKEN::NE_OP);
               }
";"            { count(YYText());
                 yylval->emplace<std::pair<int, int> >
                 (std::make_pair(Pos::linenum, Pos::colnum));
                 return(';');
               }
"{"            { count(YYText());
                 yylval->emplace<std::pair<int, int> >
                 (std::make_pair(Pos::linenum, Pos::colnum));
                 return('{');
               }
"}"            { count(YYText());
                 yylval->emplace<std::pair<int, int> >
                 (std::make_pair(Pos::linenum, Pos::colnum));
                 return('}');
               }
","            { count(YYText());
                 yylval->emplace<std::pair<int, int> >
                 (std::make_pair(Pos::linenum, Pos::colnum));
                 return(',');
               }
":"            { count(YYText());
                 yylval->emplace<std::pair<int, int> >
                 (std::make_pair(Pos::linenum, Pos::colnum));
                 return(':');
               }
"="            { count(YYText());
                 yylval->emplace<std::pair<int, int> >
                 (std::make_pair(Pos::linenum, Pos::colnum));
                 return('=');
               }
"("            { count(YYText());
                 yylval->emplace<std::pair<int, int> >
                 (std::make_pair(Pos::linenum, Pos::colnum));
                 return('(');
               }
")"            { count(YYText());
                 yylval->emplace<std::pair<int, int> >
                 (std::make_pair(Pos::linenum, Pos::colnum));
                 return(')');
               }
"["            { count(YYText());
                 yylval->emplace<std::pair<int, int> >
                 (std::make_pair(Pos::linenum, Pos::colnum));
                 return('[');
               }
"]"            { count(YYText());
                 yylval->emplace<std::pair<int, int> >
                 (std::make_pair(Pos::linenum, Pos::colnum));
                 return(']');
               }
"."            { count(YYText());
                 yylval->emplace<std::pair<int, int> >
                 (std::make_pair(Pos::linenum, Pos::colnum));
                 return('.');
               }
"&"            { count(YYText());
                 yylval->emplace<std::pair<int, int> >
                 (std::make_pair(Pos::linenum, Pos::colnum));
                 return('&');
               }
"!"            { count(YYText());
                 yylval->emplace<std::pair<int, int> >
                 (std::make_pair(Pos::linenum, Pos::colnum));
                 return('!');
               }
"~"            { count(YYText());
                 yylval->emplace<std::pair<int, int> >
                 (std::make_pair(Pos::linenum, Pos::colnum));
                 return('~');
               }
"-"            { count(YYText());
                 yylval->emplace<std::pair<int, int> >
                 (std::make_pair(Pos::linenum, Pos::colnum));
                 return('-');
               }
"+"            { count(YYText());
                 yylval->emplace<std::pair<int, int> >
                 (std::make_pair(Pos::linenum, Pos::colnum));
                 return('+');
               }
"*"            { count(YYText());
                 yylval->emplace<std::pair<int, int> >
                 (std::make_pair(Pos::linenum, Pos::colnum));
                 return('*');
               }
"/"            { count(YYText());
                 yylval->emplace<std::pair<int, int> >
                 (std::make_pair(Pos::linenum, Pos::colnum));
                 return('/');
               }
"%"            { count(YYText());
                 yylval->emplace<std::pair<int, int> >
                 (std::make_pair(Pos::linenum, Pos::colnum));
                 return('%');
               }
"<"            { count(YYText());
                 yylval->emplace<std::pair<int, int> >
                 (std::make_pair(Pos::linenum, Pos::colnum));
                 return('<');
               }
">"            { count(YYText());
                 yylval->emplace<std::pair<int, int> >
                 (std::make_pair(Pos::linenum, Pos::colnum));
                 return('>');
               }
"^"            { count(YYText());
                 yylval->emplace<std::pair<int, int> >
                 (std::make_pair(Pos::linenum, Pos::colnum));
                 return('^');
               }
"|"            { count(YYText());
                 yylval->emplace<std::pair<int, int> >
                 (std::make_pair(Pos::linenum, Pos::colnum));
                 return('|');
               }
"?"            { count(YYText());
                 yylval->emplace<std::pair<int, int> >
                 (std::make_pair(Pos::linenum, Pos::colnum));
                 return('?');
               }

[ \t]    { count(YYText()); }
[\r\n\f] { Pos::line++; Pos::col=1; Pos::setpos(); }
.        { count(YYText());
           std::string error_str = YYText();
           throw InvalidTokenException(error_str, Pos::linenum, Pos::colnum);
         }


%%

std::string quant_c_compiler::get_token_name(int value) {
    switch(value) {
        case YY_NULL:               return "EOF";
        case TOKEN::IDENTIFIER:     return "IDENTIFIER";
        case TOKEN::INT_LITERAL:    return "INT_LITERAL";
        case TOKEN::FLOAT_LITERAL:  return "FLOAT_LITERAL";
        case TOKEN::BOOL_LITERAL:   return "BOOL_LITERAL";
        case TOKEN::CHAR_LITERAL:   return "CHAR_LITERAL";
        case TOKEN::STRING_LITERAL: return "STRING_LITERAL";
        case TOKEN::SIZEOF:         return "SIZEOF";
        case TOKEN::PTR_OP:         return "PTR_OP";
        case TOKEN::INC_OP:         return "INC_OP";
        case TOKEN::DEC_OP:         return "DEC_OP";
        case TOKEN::LEFT_OP:        return "LEFT_OP";
        case TOKEN::RIGHT_OP:       return "RIGHT_OP";
        case TOKEN::LE_OP:          return "LE_OP";
        case TOKEN::GE_OP:          return "GE_OP";
        case TOKEN::EQ_OP:          return "EQ_OP";
        case TOKEN::NE_OP:          return "NE_OP";
        case TOKEN::AND_OP:         return "AND_OP";
        case TOKEN::OR_OP:          return "OR_OP";
        case TOKEN::MUL_ASSIGN:     return "MUL_ASSIGN";
        case TOKEN::DIV_ASSIGN:     return "DIV_ASSIGN";
        case TOKEN::MOD_ASSIGN:     return "MOD_ASSIGN";
        case TOKEN::ADD_ASSIGN:     return "ADD_ASSIGN";
        case TOKEN::SUB_ASSIGN:     return "SUB_ASSIGN";
        case TOKEN::LEFT_ASSIGN:    return "LEFT_ASSIGN";
        case TOKEN::RIGHT_ASSIGN:   return "RIGHT_ASSIGN";
        case TOKEN::AND_ASSIGN:     return "AND_ASSIGN";
        case TOKEN::XOR_ASSIGN:     return "XOR_ASSIGN";
        case TOKEN::OR_ASSIGN:      return "OR_ASSIGN";
        case TOKEN::TYPE_NAME:      return "TYPE_NAME";
        case TOKEN::TYPEDEF:        return "TYPEDEF";
        case TOKEN::EXTERN:         return "EXTERN";
        case TOKEN::STATIC:         return "STATIC";
        case TOKEN::AUTO:           return "AUTO";
        case TOKEN::REGISTER:       return "REGISTER";
        case TOKEN::CHAR:           return "CHAR";
        case TOKEN::SHORT:          return "SHORT";
        case TOKEN::INT:            return "INT";
        case TOKEN::LONG:           return "LONG";
        case TOKEN::SIGNED:         return "SIGNED";
        case TOKEN::UNSIGNED:       return "UNSIGNED";
        case TOKEN::FLOAT:          return "FLOAT";
        case TOKEN::DOUBLE:         return "DOUBLE";
        case TOKEN::CONST:          return "CONST";
        case TOKEN::VOLATILE:       return "VOLATILE";
        case TOKEN::VOID:           return "VOID";
        case TOKEN::STRUCT:         return "STRUCT";
        case TOKEN::UNION:          return "UNION";
        case TOKEN::ENUM:           return "ENUM";
        case TOKEN::ELLIPSIS:       return "ELLIPSIS";
        case TOKEN::QUANTUM:        return "QUANTUM";
        case TOKEN::IDENTITY:       return "IDENTITY";
        case TOKEN::MEASURE:        return "MEASURE";
        case TOKEN::PSWAP:          return "PSWAP";
        case TOKEN::ISWAP:          return "ISWAP";
        case TOKEN::SWAP:           return "SWAP";
        case TOKEN::Y_GATE:         return "Y_GATE";
        case TOKEN::Z_GATE:         return "Z_GATE";
        case TOKEN::HADAMARD:       return "HADAMARD";
        case TOKEN::PHASE:          return "PHASE";
        case TOKEN::S_GATE:         return "S_GATE";
        case TOKEN::T_GATE:         return "T_GATE";
        case TOKEN::RX:             return "RX";
        case TOKEN::RY:             return "RY";
        case TOKEN::RZ:             return "RZ";
        case TOKEN::INVERSE:        return "INVERSE";
        case TOKEN::CONTROL:        return "CONTROL";
        case TOKEN::BOOL:           return "BOOL";
        case TOKEN::CASE:           return "CASE";
        case TOKEN::DEFAULT:        return "DEFAULT";
        case TOKEN::IF:             return "IF";
        case TOKEN::ELSE:           return "ELSE";
        case TOKEN::SWITCH:         return "SWITCH";
        case TOKEN::WHILE:          return "WHILE";
        case TOKEN::DO:             return "DO";
        case TOKEN::FOR:            return "FOR";
        case TOKEN::GOTO:           return "GOTO";
        case TOKEN::CONTINUE:       return "CONTINUE";
        case TOKEN::BREAK:          return "BREAK";
        case TOKEN::RETURN:         return "RETURN";
        case ';':                   return ";";
        case '{':                   return "{";
        case '}':                   return "}";
        case ',':                   return ",";
        case ':':                   return ":";
        case '=':                   return "=";
        case '(':                   return "(";
        case ')':                   return ")";
        case '[':                   return "[";
        case ']':                   return "]";
        case '.':                   return ".";
        case '&':                   return "&";
        case '!':                   return "!";
        case '~':                   return "~";
        case '-':                   return "-";
        case '+':                   return "+";
        case '*':                   return "*";
        case '/':                   return "/";
        case '%':                   return "%";
        case '<':                   return "<";
        case '>':                   return ">";
        case '^':                   return "^";
        case '|':                   return "|";
        case '?':                   return "?";
        default: return "UNKNOWN";
    }
}

// Need to define to satisfy compiler
int yyFlexLexer::yylex() {
    throw std::runtime_error("Bad call to yyFlexLexer::yylex()");
}

void count(const char* yytext)
{
    std::string yytext_str = yytext;
    Pos::setpos();
    Pos::calcPos(yytext_str);
}

bool ends_with(std::string const &string, std::string const &phrase) {
    if (phrase.size() > string.size()) return false;

    // Look through both strings backwards to see if the given string
    // ends with the given phrase
    return std::equal(phrase.rbegin(), phrase.rend(), string.rbegin());
}

void decode_int(const char * yytext, Parser::semantic_type *yylval, int base) {
    std::string yytext_str = yytext;
    if (ends_with(yytext_str, "u") || ends_with(yytext_str, "U")) {
        IntLitToken int_token = IntLitToken(strtol(yytext, nullptr, base),
                                IntLitSize::INT,
                                false,
                                Pos::linenum,
                                Pos::colnum
                               );
        yylval->emplace<IntLitToken>(int_token);
    }
    else if (ends_with(yytext_str, "ul") || ends_with(yytext_str, "UL")) {
        IntLitToken int_token = IntLitToken(strtoul(yytext, nullptr, base),
                                IntLitSize::LONG,
                                false,
                                Pos::linenum,
                                Pos::colnum
                               );
        yylval->emplace<IntLitToken>(int_token);
    }
    else if (ends_with(yytext_str, "ull") || ends_with(yytext_str, "ULL")) {
        IntLitToken int_token = IntLitToken(strtoull(yytext, nullptr, base),
                                IntLitSize::LONGLONG,
                                false,
                                Pos::linenum,
                                Pos::colnum
                               );
        yylval->emplace<IntLitToken>(int_token);
    }
    else if (ends_with(yytext_str, "l") || ends_with(yytext_str, "L")) {
        IntLitToken int_token = IntLitToken(strtol(yytext, nullptr, base),
                                IntLitSize::LONG,
                                true,
                                Pos::linenum,
                                Pos::colnum
                               );
        yylval->emplace<IntLitToken>(int_token);
    }
    else if (ends_with(yytext_str, "ll") || ends_with(yytext_str, "LL")) {
        IntLitToken int_token = IntLitToken(strtoll(yytext, nullptr, base),
                                IntLitSize::LONGLONG,
                                true,
                                Pos::linenum,
                                Pos::colnum
                               );
        yylval->emplace<IntLitToken>(int_token);
    }
    else {
        // Process as a standard int; invalid characters will be ignored
        IntLitToken int_token = IntLitToken(strtol(yytext, nullptr, base),
                                IntLitSize::INT,
                                true,
                                Pos::linenum,
                                Pos::colnum
                               );
        yylval->emplace<IntLitToken>(int_token);
    }
}

void decode_float(const char * yytext, Parser::semantic_type *yylval) {
    std::string yytext_str = yytext;
    if (ends_with(yytext_str, "f") || ends_with(yytext_str, "F")) {
        FloatLitToken float_token = FloatLitToken(strtof(yytext, nullptr),
                                    FloatLitSize::FLOAT,
                                    Pos::linenum,
                                    Pos::colnum
                                   );
        yylval->emplace<FloatLitToken>(float_token);
    }
    else if (ends_with(yytext_str, "l") || ends_with(yytext_str, "L")) {
        FloatLitToken float_token = FloatLitToken(strtold(yytext, nullptr),
                                    FloatLitSize::LONGDOUBLE,
                                    Pos::linenum,
                                    Pos::colnum
                                   );
        yylval->emplace<FloatLitToken>(float_token);
    }
    else {
        // Process as a double; invalid characters will be ignored
        FloatLitToken float_token = FloatLitToken(strtod(yytext, nullptr),
                                    FloatLitSize::DOUBLE,
                                    Pos::linenum,
                                    Pos::colnum
                                   );
        yylval->emplace<FloatLitToken>(float_token);
    }
}
