%require "3.5"
%language "C++"

%define parse.error verbose
%define parse.trace
%define api.parser.class {Parser}
%define api.namespace {quant_c_compiler}
%define api.value.type variant
%parse-param {Scanner& scanner} {quant_c_ast::TranslationUnitNode& prog}

%code requires
{
    #include <unordered_map>
    #include <utility>
    #include "ast/ast.hpp"
    #include "token/token.hpp"

    // using namespace quant_c_ast;
    namespace quant_c_compiler {
        class Scanner;
    } // namespace quant_c_compiler
}

%code
{
    #include <iostream>
    #include <string>
    #include <cctype>
    #include <cstring>
    #include <vector>
    #include <tuple>
    #include <memory>
    #include <cstdint>


    // Adjust how Bison calls the scanner for the next token
    #include "scanner_shim.hpp"
    #define yylex(x) scanner.lex(x)
    #include "util/pos.hpp"
    #include "util/type_def_set.hpp"


    using namespace quant_c_compiler;
    using namespace quant_c_token;
    using namespace quant_c_ast;
    using namespace util;
}

/* Standar C tokens */
%token <quant_c_token::IdentifierToken> IDENTIFIER
%token <quant_c_token::IntLitToken> INT_LITERAL
%token <quant_c_token::FloatLitToken> FLOAT_LITERAL
%token <quant_c_token::BoolLitToken> BOOL_LITERAL
%token <quant_c_token::CharLitToken> CHAR_LITERAL
%token <quant_c_token::StringLitToken> STRING_LITERAL
%token <quant_c_token::TypeNameToken> TYPE_NAME
%token <std::pair<int, int>> PTR_OP INC_OP DEC_OP LEFT_OP RIGHT_OP LE_OP GE_OP
%token <std::pair<int, int>> EQ_OP NE_OP ADD_ASSIGN
%token <std::pair<int, int>> AND_OP OR_OP MUL_ASSIGN DIV_ASSIGN MOD_ASSIGN
%token <std::pair<int, int>> SUB_ASSIGN LEFT_ASSIGN RIGHT_ASSIGN AND_ASSIGN
%token <std::pair<int, int>> XOR_ASSIGN OR_ASSIGN ELLIPSIS

%token <std::pair<int, int>> ';' '{' '}' ',' ':' '=' '(' ')' '[' ']' '.' '&'
%token <std::pair<int, int>> '!' '~' '-' '+' '*' '/' '%' '<' '>' '^' '|' '?'

%token <std::pair<int, int>> TYPEDEF EXTERN STATIC AUTO REGISTER
%token <std::pair<int, int>> CHAR SHORT INT LONG SIGNED UNSIGNED FLOAT DOUBLE
%token <std::pair<int, int>> CONST VOLATILE VOID STRUCT UNION ENUM

%token <std::pair<int, int>> CASE DEFAULT IF ELSE SWITCH WHILE DO FOR
%token <std::pair<int, int>> GOTO CONTINUE BREAK RETURN SIZEOF

/* New bool type (like in C++) */
%token <std::pair<int, int>> BOOL

/* Quanutm tokens */
%token <std::pair<int, int>> QUANTUM
%token <std::pair<int, int>> PSWAP ISWAP SWAP
%token <std::pair<int, int>> X_GATE Y_GATE Z_GATE HADAMARD PHASE S_GATE T_GATE RX RY RZ
%token <std::pair<int, int>> INVERSE CONTROL
%token <std::pair<int, int>> MEASURE IDENTITY

/* Non-terminal symbols */
%nterm <std::shared_ptr<quant_c_ast::ExprNode>>  expression
%nterm <std::shared_ptr<quant_c_ast::ExprNode>>  primary_expression
%nterm <std::shared_ptr<quant_c_ast::ExprNode>>  postfix_expression
%nterm <std::shared_ptr<quant_c_ast::ExprNode>>  unary_expression
%nterm <std::shared_ptr<quant_c_ast::ExprNode>>  cast_expression
%nterm <std::shared_ptr<quant_c_ast::ExprNode>>  multiplicative_expression
%nterm <std::shared_ptr<quant_c_ast::ExprNode>>  additive_expression
%nterm <std::shared_ptr<quant_c_ast::ExprNode>>  shift_expression
%nterm <std::shared_ptr<quant_c_ast::ExprNode>>  relational_expression
%nterm <std::shared_ptr<quant_c_ast::ExprNode>>  quantum_binary_expression
%nterm <std::shared_ptr<quant_c_ast::ExprNode>>  equality_expression
%nterm <std::shared_ptr<quant_c_ast::ExprNode>>  and_expression
%nterm <std::shared_ptr<quant_c_ast::ExprNode>>  exclusive_or_expression
%nterm <std::shared_ptr<quant_c_ast::ExprNode>>  inclusive_or_expression
%nterm <std::shared_ptr<quant_c_ast::ExprNode>>  logical_and_expression
%nterm <std::shared_ptr<quant_c_ast::ExprNode>>  logical_or_expression
%nterm <std::shared_ptr<quant_c_ast::ExprNode>>  conditional_expression
%nterm <std::shared_ptr<quant_c_ast::ExprNode>>  assignment_expression
%nterm <std::shared_ptr<quant_c_ast::ExprNode>>  constant_expression
%nterm <std::shared_ptr<quant_c_ast::ExprNode>>  initializer


%nterm <std::vector<std::shared_ptr<quant_c_ast::ExprNode>>> argument_expression_list
%nterm <std::vector<std::shared_ptr<quant_c_ast::ExprNode>>> initializer_list

%nterm <std::shared_ptr<quant_c_ast::DeclNode>> declaration
%nterm <std::shared_ptr<quant_c_ast::DeclaratorNode>> declarator
%nterm <std::shared_ptr<quant_c_ast::DirectDeclaratorNode>> direct_declarator
%nterm <std::shared_ptr<quant_c_ast::InitDeclaratorNode>> init_declarator
// %nterm <std::shared_ptr<quant_c_ast::DeclNode>> struct_declaration
%nterm <std::shared_ptr<quant_c_ast::ArgDeclNode>> parameter_declaration
%nterm <std::shared_ptr<quant_c_ast::DeclNode>> external_declaration

%nterm <std::vector<std::shared_ptr<quant_c_ast::InitDeclaratorNode>>> init_declarator_list
// %nterm <std::vector<std::shared_ptr<quant_c_ast::DeclNode>>> struct_declaration_list
%nterm <std::vector<std::shared_ptr<quant_c_ast::DeclNode>>> declaration_list
%nterm <std::pair<std::vector<std::shared_ptr<quant_c_ast::ArgDeclNode>>, bool>> parameter_type_list
%nterm <std::vector<std::shared_ptr<quant_c_ast::ArgDeclNode>>> parameter_list


%nterm <std::shared_ptr<quant_c_ast::TypeSpecNode>> type_specifier
%nterm <std::shared_ptr<quant_c_ast::TypeQualifierNode>> type_qualifier
%nterm <std::shared_ptr<quant_c_ast::StorageClassSpecNode>> storage_class_specifier

// %nterm <std::shared_ptr<quant_c_ast::TypeNode>> struct_or_union_specifier
// %nterm <std::shared_ptr<quant_c_ast::TypeNode>> enum_specifier
// %nterm <std::shared_ptr<quant_c_ast::TypeNode>> struct_declarator enumerator
%nterm <std::shared_ptr<quant_c_ast::TypeNameNode>> type_name
%nterm <std::shared_ptr<quant_c_ast::AbstractDeclaratorNode>> abstract_declarator
%nterm <std::shared_ptr<quant_c_ast::DirectAbstractDeclaratorNode>> direct_abstract_declarator

%nterm <std::vector<std::shared_ptr<quant_c_ast::TypeNode>>> declaration_specifiers
%nterm <std::vector<std::shared_ptr<quant_c_ast::PointerNode>>> pointer
%nterm <std::vector<std::shared_ptr<quant_c_ast::TypeNode>>> specifier_qualifier_list
// %nterm <std::vector<std::shared_ptr<quant_c_ast::TypeNode>>> struct_declarator_list
// %nterm <std::vector<std::shared_ptr<quant_c_ast::TypeNode>>> enumerator_list
%nterm <std::vector<std::shared_ptr<quant_c_ast::TypeQualifierNode>>> type_qualifier_list

%nterm <std::shared_ptr<quant_c_ast::StmtNode>> statement labeled_statement
%nterm <std::shared_ptr<quant_c_ast::BlockNode>> compound_statement
%nterm <std::shared_ptr<quant_c_ast::StmtNode>> expression_statement
%nterm <std::shared_ptr<quant_c_ast::StmtNode>> selection_statement
%nterm <std::shared_ptr<quant_c_ast::StmtNode>> iteration_statement
%nterm <std::shared_ptr<quant_c_ast::StmtNode>> jump_statement

%nterm <std::vector<std::shared_ptr<quant_c_ast::StmtNode>>> statement_list

%nterm <std::shared_ptr<quant_c_ast::FuncDefNode>> function_definition

%nterm <std::shared_ptr<quant_c_ast::OperatorNode>> unary_operator
%nterm <int> assignment_operator
%nterm <std::tuple<quant_c_ast::QuantumUnaryOperatorNode::QuantumUnaryOp, int, int>> quantum_unary_op

%nterm <std::tuple<bool, int, int>> quantum_op_prefix
// %nterm <int> struct_or_union

%nterm <std::vector<std::shared_ptr<quant_c_ast::QuantumUnaryPostfixNode>>> quantum_op_postfix

%nterm <std::vector<std::shared_ptr<quant_c_token::IdentifierToken>>> identifier_list

%start translation_unit

%%

primary_expression
	: IDENTIFIER {$$ = std::make_shared<IdentNode>($1.get_value(),
                                              $1.get_line(),
                                              $1.get_col()); }
	| INT_LITERAL    { $$ = std::make_shared<IntLitNode>($1.get_value(),
                                                         $1.get_size(),
                                                         $1.is_signed(),
                                                         $1.get_line(),
                                                         $1.get_col());
                    }
    | FLOAT_LITERAL { $$ = std::make_shared<FloatLitNode>($1.get_value(),
                                                          $1.get_size(),
                                                          $1.get_line(),
                                                          $1.get_col());;
                    }
    | BOOL_LITERAL {
                     if ($1.get_value()) {
                        $$ = std::make_shared<TrueNode>($1.get_line(), $1.get_col());
                     } else {
                        $$ = std::make_shared<FalseNode>($1.get_line(), $1.get_col());
                     }
    }
    | CHAR_LITERAL  {$$ = std::make_shared<CharLitNode>($1.get_value(),
                                                   $1.get_line(),
                                                   $1.get_col());
                    }
	| STRING_LITERAL {$$ = std::make_shared<StrLitNode>($1.get_value(),
                                                    $1.get_value(false),
                                                    $1.get_line(),
                                                    $1.get_col());
                    }
	| '(' expression ')' {$$ = $2; }
	;

postfix_expression
	: primary_expression                       { $$ = $1; }
	| postfix_expression '[' expression ']'    {
        $$ = std::make_shared<IndexExprNode>($1, $3, $1->get_line(), $1->get_col());
    }
	| postfix_expression '(' ')'
    {
        auto no_args = std::vector<std::shared_ptr<ExprNode>>();
        $$ = std::make_shared<CallExprNode>($1, no_args,
                                            $1->get_line(),
                                            $1->get_col());
    }
	| postfix_expression '(' argument_expression_list ')'
    {
        $$ = std::make_shared<CallExprNode>($1, $3,
                                            $1->get_line(),
                                            $1->get_col());
    }
	| postfix_expression '.' IDENTIFIER
	| postfix_expression PTR_OP IDENTIFIER
	| postfix_expression INC_OP {
        auto op = std::make_shared<UnaryOperatorNode>(Parser::token_type::INC_OP,
                                                true,
                                                $2.first, $2.second);
        $$ = std::make_shared<UnaryOpNode>(op, $1,
                                           $1->get_line(), $1->get_col());
    }
	| postfix_expression DEC_OP {
        auto op = std::make_shared<UnaryOperatorNode>(Parser::token_type::DEC_OP,
                                                true,
                                                $2.first, $2.second);
        $$ = std::make_shared<UnaryOpNode>(op, $1,
                                           $1->get_line(), $1->get_col());
    }
	;

argument_expression_list
	: assignment_expression
    {
        $$ = std::vector<std::shared_ptr<ExprNode>>();
        $$.push_back($1);
    }
	| argument_expression_list ',' assignment_expression
    {
        $$ = std::vector<std::shared_ptr<ExprNode>>($1);
        $$.push_back($3);
    }
	;

unary_expression
	: postfix_expression { $$ = $1; }
	| INC_OP unary_expression {
        auto op =
            std::make_shared<UnaryOperatorNode>(Parser::token_type::INC_OP,
                                                false,
                                                $1.first, $1.second);
        $$ = std::make_shared<UnaryOpNode>(op, $2, $1.first, $1.second);
    }
	| DEC_OP unary_expression {
        auto op =
            std::make_shared<UnaryOperatorNode>(Parser::token_type::DEC_OP,
                                                false,
                                                $1.first, $1.second);
        $$ = std::make_shared<UnaryOpNode>(op, $2, $1.first, $1.second);
    }
	| unary_operator cast_expression
    {
        $$ = std::make_shared<UnaryOpNode>($1, $2,
                                           $1->get_line(), $1->get_col());
    }
	| SIZEOF unary_expression
    //TODO: change
    { $$ = $2; }
	| SIZEOF '(' type_name ')' { $$ = ExprNode::NULL_VAL; }
	;

unary_operator
	: '&' {
        $$ = std::make_shared<UnaryOperatorNode>('&');
    }
	| '*' {
        $$ = std::make_shared<UnaryOperatorNode>('*');
    }
	| '+' {
        $$ = std::make_shared<UnaryOperatorNode>('+');
    }
	| '-' {
        $$ = std::make_shared<UnaryOperatorNode>('-');
    }
	| '~' {
        $$ = std::make_shared<UnaryOperatorNode>('~');
    }
	| '!' {
        $$ = std::make_shared<UnaryOperatorNode>('!');
    }
    | quantum_op_prefix quantum_unary_op quantum_op_postfix {
        // Determine line/col info
        auto isInverted = std::get<0>($1);
        auto opCode = std::get<0>($2);
        const int line_index = 1;
        const int col_index = 2;
        int first = std::get<line_index>($1);
        if (first == -1) { first = std::get<line_index>($2); }
        int second = std::get<col_index>($1);
        if (second == -1) { second = std::get<col_index>($2); }


        $$ = std::make_shared<QuantumUnaryOperatorNode>(isInverted, opCode, $3,
                                                        first, second);
    }
	;

// NEW: Quantum operations with operation modifiers
// TODO: handle phase info
quantum_unary_op
    : IDENTITY
    {
        $$ = std::make_tuple(QuantumUnaryOperatorNode::QuantumUnaryOp::IDENTITY,
                             $1.first,
                             $1.second);
    }
    | X_GATE
    {
        $$ = std::make_tuple(QuantumUnaryOperatorNode::QuantumUnaryOp::X_GATE,
                             $1.first,
                             $1.second);
    }
    | Y_GATE
    {
        $$ = std::make_tuple(QuantumUnaryOperatorNode::QuantumUnaryOp::Y_GATE,
                             $1.first,
                             $1.second);
    }
    | Z_GATE
    {
        $$ = std::make_tuple(QuantumUnaryOperatorNode::QuantumUnaryOp::Z_GATE,
                             $1.first,
                             $1.second);
    }
    | HADAMARD
    {
        $$ = std::make_tuple(QuantumUnaryOperatorNode::QuantumUnaryOp::HADAMARD,
                             $1.first,
                             $1.second);
    }
    | PHASE '(' FLOAT_LITERAL ')'
    {
        $$ = std::make_tuple(QuantumUnaryOperatorNode::QuantumUnaryOp::PHASE,
                             $1.first,
                             $1.second);
    }
    | S_GATE
    {
        $$ = std::make_tuple(QuantumUnaryOperatorNode::QuantumUnaryOp::S_GATE,
                             $1.first,
                             $1.second);
    }
    | T_GATE
    {
        $$ = std::make_tuple(QuantumUnaryOperatorNode::QuantumUnaryOp::T_GATE,
                             $1.first,
                             $1.second);
    }
    | RX '(' FLOAT_LITERAL ')'
    {
        $$ = std::make_tuple(QuantumUnaryOperatorNode::QuantumUnaryOp::RX,
                             $1.first,
                             $1.second);
    }
    | RY '(' FLOAT_LITERAL ')'
    {
        $$ = std::make_tuple(QuantumUnaryOperatorNode::QuantumUnaryOp::RY,
                             $1.first,
                             $1.second);
    }
    | RZ '(' FLOAT_LITERAL ')'
    {
        $$ = std::make_tuple(QuantumUnaryOperatorNode::QuantumUnaryOp::RZ,
                             $1.first,
                             $1.second);
    }
    | MEASURE
    {
        $$ = std::make_tuple(QuantumUnaryOperatorNode::QuantumUnaryOp::MEASURE,
                             $1.first,
                             $1.second);
    }
    ;

quantum_op_prefix
    : %empty { $$ = std::make_tuple(false, -1, -1); }
    | INVERSE { $$ = std::make_tuple(true, $1.first, $1.second); }
    ;

quantum_op_postfix
    : %empty
    {
        $$ = std::vector<std::shared_ptr<QuantumUnaryPostfixNode>>();
    }
    | CONTROL '(' IDENTIFIER ')' quantum_op_postfix
    {
        auto controlList = std::vector<std::shared_ptr<QuantumUnaryPostfixNode>>($5);
        auto ident = IdentNode($3.get_value(),
                               $3.get_line(),
                               $3.get_col());
        controlList.insert(controlList.begin(),
                           std::make_shared<QuantumUnaryPostfixNode>(
                                                            ident,
                                                            $1.first,
                                                            $1.second));
        $$ = controlList;

    }
    ;

cast_expression
	: unary_expression { $$ = $1; }
	| '(' type_name ')' cast_expression {
        $$ = std::make_shared<CastNode>($2, $4, $1.first, $1.second);
    }
	;

multiplicative_expression
	: cast_expression { $$ = $1; }
	| multiplicative_expression '*' cast_expression {
        $$ = std::make_shared<BinaryOpNode>($1, '*', $3, $1->get_line(), $1->get_col());
    }
	| multiplicative_expression '/' cast_expression {
        $$ = std::make_shared<BinaryOpNode>($1, '/', $3, $1->get_line(), $1->get_col());
    }
	| multiplicative_expression '%' cast_expression {
        $$ = std::make_shared<BinaryOpNode>($1, '%', $3, $1->get_line(), $1->get_col());
    }
	;

additive_expression
	: multiplicative_expression { $$ = $1; }
	| additive_expression '+' multiplicative_expression {
        $$ = std::make_shared<BinaryOpNode>($1, '+', $3, $1->get_line(), $1->get_col());
    }
	| additive_expression '-' multiplicative_expression {
        $$ = std::make_shared<BinaryOpNode>($1, '-', $3, $1->get_line(), $1->get_col());
    }
	;

shift_expression
	: additive_expression { $$ = $1; }
	| shift_expression LEFT_OP additive_expression {
        $$ = std::make_shared<BinaryOpNode>($1, Parser::token_type::LEFT_OP, $3,
                                            $1->get_line(), $1->get_col());
    }
	| shift_expression RIGHT_OP additive_expression {
        $$ = std::make_shared<BinaryOpNode>($1,
                                            Parser::token_type::RIGHT_OP, $3,
                                            $1->get_line(), $1->get_col());
    }
	;

relational_expression
	: shift_expression { $$ = $1; }
	| relational_expression '<' shift_expression {
        $$ = std::make_shared<BinaryOpNode>($1, '<', $3, $1->get_line(), $1->get_col());
    }
	| relational_expression '>' shift_expression {
        $$ = std::make_shared<BinaryOpNode>($1, '>', $3, $1->get_line(), $1->get_col());
    }
	| relational_expression LE_OP shift_expression {
        $$ = std::make_shared<BinaryOpNode>($1, Parser::token_type::LE_OP, $3,
                                            $1->get_line(), $1->get_col());
    }
	| relational_expression GE_OP shift_expression {
        $$ = std::make_shared<BinaryOpNode>($1, Parser::token_type::GE_OP, $3,
                                            $1->get_line(), $1->get_col());
    }
    | quantum_binary_expression { $$ = $1; }
	;

quantum_binary_expression
    : relational_expression PSWAP '(' expression ')' shift_expression
    | relational_expression SWAP shift_expression
    | relational_expression ISWAP shift_expression
    ;

equality_expression
	: relational_expression { $$ = $1; }
	| equality_expression EQ_OP relational_expression {
        $$ = std::make_shared<BinaryOpNode>($1, Parser::token_type::EQ_OP, $3,
                                            $1->get_line(), $1->get_col());
    }
	| equality_expression NE_OP relational_expression {
        $$ = std::make_shared<BinaryOpNode>($1, Parser::token_type::NE_OP, $3,
                                            $1->get_line(), $1->get_col());
    }
	;

and_expression
	: equality_expression { $$ = $1; }
	| and_expression '&' equality_expression {
        $$ = std::make_shared<BinaryOpNode>($1, '&', $3, $1->get_line(), $1->get_col());
    }
    // NEW: Trinary version of AND (intended for quantum variables)
    | and_expression '&' equality_expression ':' IDENTIFIER
	;

exclusive_or_expression
	: and_expression     { $$ = $1; }
	| exclusive_or_expression '^' and_expression
    {
        $$ = std::make_shared<BinaryOpNode>($1, '^', $3,
                                            $1->get_line(),
                                            $1->get_col());
    }
	;

inclusive_or_expression
	: exclusive_or_expression   { $$ = $1; }
	| inclusive_or_expression '|' exclusive_or_expression {
        $$ = std::make_shared<BinaryOpNode>($1, '|', $3,
                                            $1->get_line(), $1->get_col());
    }
    // NEW: Trinary version of OR (intended for quantum variables)
    | inclusive_or_expression '|' exclusive_or_expression ':' IDENTIFIER
	;

logical_and_expression
	: inclusive_or_expression   { $$ = $1; }
	| logical_and_expression AND_OP inclusive_or_expression {
        $$ = std::make_shared<BinaryOpNode>($1, Parser::token_type::AND_OP, $3,
                                            $1->get_line(), $1->get_col());
    }
	;

logical_or_expression
	: logical_and_expression    { $$ = $1; }
	| logical_or_expression OR_OP logical_and_expression {
        $$ = std::make_shared<BinaryOpNode>($1, Parser::token_type::OR_OP, $3,
                                            $1->get_line(), $1->get_col());
    }
	;

conditional_expression
	: logical_or_expression  { $$ = $1; }
	| logical_or_expression '?' expression ':' conditional_expression {
        $$ = std::make_shared<ConditionalExprNode>($1, $3, $5,
                                                  $1->get_line(),
                                                  $1->get_col());
    }
	;

assignment_expression
	: conditional_expression { $$ = $1; }
	| unary_expression assignment_operator assignment_expression {
        $$ = std::make_shared<AsgExprNode>($1, $2, $3,
                                           $1->get_line(), $1->get_col());
    }
	;

assignment_operator
	: '='           { $$ = (int)'='; }
	| MUL_ASSIGN    { $$ = Parser::token_type::MUL_ASSIGN; }
	| DIV_ASSIGN    { $$ = Parser::token_type::DIV_ASSIGN; }
	| MOD_ASSIGN    { $$ = Parser::token_type::MOD_ASSIGN; }
	| ADD_ASSIGN    { $$ = Parser::token_type::ADD_ASSIGN; }
	| SUB_ASSIGN    { $$ = Parser::token_type::SUB_ASSIGN; }
	| LEFT_ASSIGN   { $$ = Parser::token_type::LEFT_ASSIGN; }
	| RIGHT_ASSIGN  { $$ = Parser::token_type::RIGHT_ASSIGN; }
	| AND_ASSIGN    { $$ = Parser::token_type::AND_ASSIGN; }
	| XOR_ASSIGN    { $$ = Parser::token_type::XOR_ASSIGN; }
	| OR_ASSIGN     { $$ = Parser::token_type::OR_ASSIGN; }
	;

expression
	: assignment_expression {$$ = $1;}
	| expression ',' assignment_expression
    {
        $$ = std::make_shared<BinaryOpNode>($1, ',', $3,
                                            $1->get_line(),
                                            $1->get_col());
    }
	;

constant_expression
	: conditional_expression  { $$ = $1; }
	;

declaration
	: declaration_specifiers ';'
    {
        auto no_init = std::vector<std::shared_ptr<InitDeclaratorNode>>();
        $$ = std::make_shared<DeclarationNode>($1,
                                               no_init,
                                               $1.front()->get_line(),
                                               $1.front()->get_col());
    }
	| declaration_specifiers init_declarator_list ';'
    {
        $$ = std::make_shared<DeclarationNode>($1,
                                               $2,
                                               $1.front()->get_line(),
                                               $1.front()->get_col());
    }
	;

declaration_specifiers
	: storage_class_specifier   {
        $$ = std::vector<std::shared_ptr<TypeNode>>();
        $$.insert($$.begin(), $1);
    }
	| storage_class_specifier declaration_specifiers {
        $$ = std::vector<std::shared_ptr<TypeNode>>($2);
        $$.insert($$.begin(), $1);
    }
	| type_specifier {
        $$ = std::vector<std::shared_ptr<TypeNode>>();
        $$.insert($$.begin(), $1);
    }
	| type_specifier declaration_specifiers {
        $$ = std::vector<std::shared_ptr<TypeNode>>($2);
        $$.insert($$.begin(), $1);
    }
	| type_qualifier {
        $$ = std::vector<std::shared_ptr<TypeNode>>();
        $$.insert($$.begin(), $1);
    }
	| type_qualifier declaration_specifiers {
        $$ = std::vector<std::shared_ptr<TypeNode>>($2);
        $$.insert($$.begin(), $1);
    }
	;

init_declarator_list
	: init_declarator
    {
        $$ = std::vector<std::shared_ptr<InitDeclaratorNode>>();
        $$.push_back($1);
    }
	| init_declarator_list ',' init_declarator
    {
        $$ = std::vector<std::shared_ptr<InitDeclaratorNode>>($1);
        $$.push_back($3);
    }
	;

init_declarator
	: declarator
    {
        $$ = std::make_shared<InitDeclaratorNode>($1,
                                                  $1->get_line(),
                                                  $1->get_col());
    }
	| declarator '=' initializer
    {
        $$ = std::make_shared<InitDeclaratorNode>($1, $3,
                                                  $1->get_line(),
                                                  $1->get_col());
    }
	;

// NEW: "quantum" specifier to denote quantum functions and variables
storage_class_specifier
	: TYPEDEF  { $$ = std::make_shared<TypeDefNode>  ($1.first, $1.second); }
	| EXTERN   { $$ = std::make_shared<ExternNode>   ($1.first, $1.second); }
	| STATIC   { $$ = std::make_shared<StaticNode>   ($1.first, $1.second); }
	| AUTO     { $$ = std::make_shared<AutoNode>     ($1.first, $1.second); }
	| REGISTER { $$ = std::make_shared<RegisterNode> ($1.first, $1.second); }
    | QUANTUM  { $$ = std::make_shared<QuantumNode>  ($1.first, $1.second); }
	;

type_specifier
	: VOID     { $$ = std::make_shared<VoidTypeNode>    ($1.first, $1.second); }
    | BOOL     { $$ = std::make_shared<BoolTypeNode>    ($1.first, $1.second); }
	| CHAR     { $$ = std::make_shared<CharTypeNode>    ($1.first, $1.second); }
	| SHORT    { $$ = std::make_shared<ShortTypeNode>   ($1.first, $1.second); }
	| INT      { $$ = std::make_shared<IntTypeNode>     ($1.first, $1.second); }
	| LONG     { $$ = std::make_shared<LongTypeNode>    ($1.first, $1.second); }
	| FLOAT    { $$ = std::make_shared<FloatTypeNode>   ($1.first, $1.second); }
	| DOUBLE   { $$ = std::make_shared<DoubleTypeNode>  ($1.first, $1.second); }
	| SIGNED   { $$ = std::make_shared<SignedTypeNode>  ($1.first, $1.second); }
	| UNSIGNED { $$ = std::make_shared<UnsignedTypeNode>($1.first, $1.second); }
	| struct_or_union_specifier {
        $$ = TypeSpecNode::NULL_VAL;
        error("structs & union are not implemented.");
    }
	| enum_specifier {
        $$ = TypeSpecNode::NULL_VAL;
        error("enum is not implemented.");
    }
	| TYPE_NAME {
        $$ = std::make_shared<CustomTypeNode>($1.get_value(),
                                              $1.get_line(),
                                              $1.get_col());
    }
	;

struct_or_union_specifier
	: struct_or_union IDENTIFIER '{' struct_declaration_list '}'
	| struct_or_union '{' struct_declaration_list '}'
	| struct_or_union IDENTIFIER
	;

struct_or_union
	: STRUCT
	| UNION
	;

struct_declaration_list
	: struct_declaration
	| struct_declaration_list struct_declaration
	;

struct_declaration
	: specifier_qualifier_list struct_declarator_list ';'
	;

specifier_qualifier_list
	: type_specifier specifier_qualifier_list {
        $$ = std::vector<std::shared_ptr<TypeNode>>($2);
        $$.insert($$.begin(), $1);
    }
	| type_specifier {
         $$ = std::vector<std::shared_ptr<TypeNode>>();
         $$.insert($$.begin(), $1);
     }
	| type_qualifier specifier_qualifier_list {
        $$ = std::vector<std::shared_ptr<TypeNode>>($2);
        $$.insert($$.begin(), $1);
    }
	| type_qualifier {
        $$ = std::vector<std::shared_ptr<TypeNode>>();
        $$.insert($$.begin(), $1);
    }
	;

struct_declarator_list
	: struct_declarator
	| struct_declarator_list ',' struct_declarator
	;

struct_declarator
	: declarator
	| ':' constant_expression
	| declarator ':' constant_expression
	;

enum_specifier
	: ENUM '{' enumerator_list '}'
	| ENUM IDENTIFIER '{' enumerator_list '}'
	| ENUM IDENTIFIER
	;

enumerator_list
	: enumerator
	| enumerator_list ',' enumerator
	;

enumerator
	: IDENTIFIER
	| IDENTIFIER '=' constant_expression
	;

type_qualifier
	: CONST     { $$ = std::make_shared<ConstNode>($1.first, $1.second); }
	| VOLATILE  { $$ = std::make_shared<VolatileNode>($1.first, $1.second); }
	;

declarator
	: pointer direct_declarator {
        $$ = std::make_shared<DeclaratorNode>($1, $2,
                                              $1.front()->get_line(),
                                              $1.front()->get_col());
    }
	| direct_declarator {
        auto non_pointer_type = std::vector<std::shared_ptr<PointerNode>>();
        $$ = std::make_shared<DeclaratorNode>(non_pointer_type, $1,
                                              $1->get_line(),
                                              $1->get_col());
    }
	;

direct_declarator
	: IDENTIFIER {
            $$ = std::make_shared<IdentifierDeclaratorNode>($1.get_value(),
                                             $1.get_line(),
                                             $1.get_col());
    }
    |
    '&' IDENTIFIER {
            $$ = std::make_shared<IdentifierDeclaratorNode>($2.get_value(),
                                             $2.get_line(),
                                             $2.get_col());
    }
	| '(' declarator ')' {
        $$ = std::make_shared<ParensDeclaratorNode>($2, $1.first, $1.second);
    }
	| direct_declarator '[' constant_expression ']' {
        $$ = std::make_shared<ArrayDeclaratorNode>($1, $3,
                                                   $1->get_line(),
                                                   $1->get_col());
    }
	| direct_declarator '[' ']' {
        $$ = std::make_shared<ArrayDeclaratorNode>($1, ExprNode::NULL_VAL,
                                                   $1->get_line(),
                                                   $1->get_col());
    }
	| direct_declarator '(' parameter_type_list ')' {
        $$ = std::make_shared<FunctionDeclaratorNode>($1, $3.first,
                                                      $3.second,
                                                      $1->get_line(),
                                                      $1->get_col());
    }
	| direct_declarator '(' identifier_list ')' {
        $$ = nullptr;
        error("Function definitions using initializer lists (K&R-style) is currently not implemented");
    }
	| direct_declarator '(' ')' {
        $$ = std::make_shared<FunctionDeclaratorNode>($1,
                                                      FunctionDeclaratorNode::
                                                      ArgumentList(),
                                                      false,
                                                      $1->get_line(),
                                                      $1->get_col());
    }
    | AUTO '[' identifier_list ']' {
        $$ = std::make_shared<DestructureDeclaratorNode>($3,
                                                         $1.first,
                                                         $1.second);
    }
	;

pointer
	: '*'
    {
        $$ = std::vector<std::shared_ptr<PointerNode>>();
        auto pointer = std::make_shared<PointerNode>(PointerNode::TypeQualifierList(),
                                      $1.first,
                                      $1.second);
        $$.insert($$.begin(), pointer);
    }
	| '*' type_qualifier_list
    {
        $$ = std::vector<std::shared_ptr<PointerNode>>();
        auto pointer = std::make_shared<PointerNode>($2,
                                                     $1.first,
                                                     $1.second);
        $$.insert($$.begin(), pointer);
    }
	| '*' pointer
    {
        $$ = std::vector<std::shared_ptr<PointerNode>>($2);
        auto pointer = std::make_shared<PointerNode>(PointerNode::TypeQualifierList(),
                                      $1.first,
                                      $1.second);
        $$.insert($$.begin(), pointer);
    }
	| '*' type_qualifier_list pointer
    {
        $$ = std::vector<std::shared_ptr<PointerNode>>($3);
        auto pointer = std::make_shared<PointerNode>($2,
                                                     $1.first,
                                                     $1.second);
        $$.insert($$.begin(), pointer);
    }
	;

type_qualifier_list
	: type_qualifier {
        $$ = std::vector<std::shared_ptr<TypeQualifierNode>>();
        $$.push_back($1);
    }
	| type_qualifier_list type_qualifier {
        $$ = std::vector<std::shared_ptr<TypeQualifierNode>>($1);
        $$.push_back($2);
    }
	;


parameter_type_list
	: parameter_list
    {
        $$ = std::make_pair($1, false);
    }
	| parameter_list ',' ELLIPSIS {
        $$ = std::make_pair($1, true);
    }
	;

parameter_list
	: parameter_declaration {
        $$ = std::vector<std::shared_ptr<ArgDeclNode>>();
        $$.push_back($1);
    }
	| parameter_list ',' parameter_declaration {
        $$ = std::vector<std::shared_ptr<ArgDeclNode>>($1);
        $$.push_back($3);
    }
	;

parameter_declaration
	: declaration_specifiers declarator {
        $$ = std::make_shared<NamedArgDeclNode>($1, $2,
                                                $1.front()->get_line(),
                                                $1.front()->get_col());
    }
	| declaration_specifiers abstract_declarator {
        $$ = std::make_shared<AbstractArgDeclNode>($1, $2,
                                                $1.front()->get_line(),
                                                $1.front()->get_col());
    }
	| declaration_specifiers {
        $$ = std::make_shared<TypeArgDeclNode>($1,
                                                $1.front()->get_line(),
                                                $1.front()->get_col());
    }
	;

identifier_list
	: IDENTIFIER {
        $$ = std::vector<std::shared_ptr<quant_c_token::IdentifierToken>>();
        auto idPtr = std::make_shared<quant_c_token::IdentifierToken>($1);
        $$.push_back(idPtr);
    }
	| identifier_list ',' IDENTIFIER {
        $$ = std::vector<std::shared_ptr<quant_c_token::IdentifierToken>>($1);
        auto idPtr = std::make_shared<quant_c_token::IdentifierToken>($3);
        $$.push_back(idPtr);
    }
	;

type_name
	: specifier_qualifier_list {
        auto frontElement = $1.front();
        $$ = std::make_shared<TypeNameNode>($1,
                                            AbstractDeclaratorNode::NULL_VAL,
                                            frontElement->get_line(), frontElement->get_col());
    }
	| specifier_qualifier_list abstract_declarator {
        auto frontElement = $1.front();
        $$ = std::make_shared<TypeNameNode>($1, $2,
                                            frontElement->get_line(),
                                            frontElement->get_col());
    }
	;

abstract_declarator
	: pointer {
        // TODO: create NULL_VAL for direct direct_abstract_declarator
        auto frontElement = $1.front();
        $$ = std::make_shared<AbstractDeclaratorNode>($1,
                                        DirectAbstractDeclaratorNode::NULL_VAL,
                                                      frontElement->get_line(),
                                                      frontElement->get_col());
    }
	| direct_abstract_declarator {
        auto emptyPointer =
                    std::vector<std::shared_ptr<quant_c_ast::PointerNode>>();
        $$ = std::make_shared<AbstractDeclaratorNode>(emptyPointer, $1,
                                                      $1->get_line(),
                                                      $1->get_col());
    }
	| pointer direct_abstract_declarator {
        auto frontElement = $1.front();
        $$ = std::make_shared<AbstractDeclaratorNode>($1, $2,
                                                      frontElement->get_line(),
                                                      frontElement->get_col());
    }
	;

direct_abstract_declarator
	: '(' abstract_declarator ')'
                                { $$ = DirectAbstractDeclaratorNode::NULL_VAL; }
	| '[' ']'                   { $$ = DirectAbstractDeclaratorNode::NULL_VAL; }
	| '[' constant_expression ']'
                                { $$ = DirectAbstractDeclaratorNode::NULL_VAL; }
	| direct_abstract_declarator '[' ']'
                                { $$ = DirectAbstractDeclaratorNode::NULL_VAL; }
	| direct_abstract_declarator '[' constant_expression ']'
                                { $$ = DirectAbstractDeclaratorNode::NULL_VAL; }
	| '(' ')'
                                { $$ = DirectAbstractDeclaratorNode::NULL_VAL; }
	| '(' parameter_type_list ')'
                                { $$ = DirectAbstractDeclaratorNode::NULL_VAL; }
	| direct_abstract_declarator '(' ')'
                                { $$ = DirectAbstractDeclaratorNode::NULL_VAL; }
	| direct_abstract_declarator '(' parameter_type_list ')'
                                { $$ = DirectAbstractDeclaratorNode::NULL_VAL; }
	;

initializer
	: assignment_expression        { $$ = $1; }
	| '{' initializer_list '}'
    {
        $$ = std::make_shared<InitializerListNode>($2,
                                                 $2.front()->get_line(),
                                                 $2.front()->get_col());
    }
	| '{' initializer_list ',' '}'
    {
        $$ = std::make_shared<InitializerListNode>($2,
                                                 $2.front()->get_line(),
                                                 $2.front()->get_col());
    }
	;

initializer_list
	: initializer
    {
        $$ = std::vector<std::shared_ptr<ExprNode>>();
        $$.push_back($1);
    }
	| initializer_list ',' initializer
    {
        $$ = std::vector<std::shared_ptr<ExprNode>>($1);
        $$.push_back($3);
    }
	;

statement
	: labeled_statement      { $$ = $1; }
	| compound_statement     { $$ = $1; }
	| expression_statement   { $$ = $1; }
	| selection_statement    { $$ = $1; }
	| iteration_statement    { $$ = $1; }
	| jump_statement         { $$ = $1; }
	;

labeled_statement
	: IDENTIFIER ':' statement {
        auto label = IdentNode($1.get_value(), $1.get_line(), $1.get_col());
        $$ = std::make_shared<LabelStmtNode>(label, $3,
                                             $1.get_line(), $1.get_col());
    }
	| CASE constant_expression ':' statement {
        $$ = std::make_shared<CaseStmtNode>($2, $4,
                                            $1.first, $1.second);
    }
	| DEFAULT ':' statement {} {
        $$ = std::make_shared<CaseStmtNode>($3,
                                            $1.first, $1.second);
    }
	;

compound_statement
	: '{' '}'                                  { $$ = BlockNode::NULL_VAL; }
	| '{' statement_list '}'                   {
        $$ = std::make_shared<BlockNode>(std::vector<std::shared_ptr<DeclNode>>(), $2, $1.first, $1.second);
    }
	| '{' declaration_list '}'                 {
        $$ = std::make_shared<BlockNode>($2, std::vector<std::shared_ptr<StmtNode>>(), $1.first, $1.second);
    }
	| '{' declaration_list statement_list '}'  {
        $$ = std::make_shared<BlockNode>($2, $3, $1.first, $1.second);
    }
	;

declaration_list
	: declaration {
        $$ = std::vector<std::shared_ptr<DeclNode>>();
        $$.push_back($1);
    }
	| declaration_list declaration  {
        $$ = std::vector<std::shared_ptr<DeclNode>>($1);
        $$.push_back($2);
    }
	;

statement_list
	: statement {
        auto new_stmt_list = std::vector<std::shared_ptr<StmtNode>>();
        new_stmt_list.push_back($1);
        $$ = new_stmt_list;
    }
	| statement_list statement {
        auto new_stmt_list = std::vector<std::shared_ptr<StmtNode>>($1);
        new_stmt_list.push_back($2);
        $$ = new_stmt_list;
    }
	;

expression_statement
	: ';'            { $$ = StmtNode::NULL_VAL; }
	| expression ';' {
        $$ = std::make_shared<ExprStmtNode>($1, $1->get_line(), $1->get_col());
      }
	;

selection_statement
	: IF '(' expression ')' statement {
        $$ = std::make_shared<IfThenNode>($3, $5, StmtNode::NULL_VAL,
                                     $1.first, $1.second);
    }
	| IF '(' expression ')' statement ELSE statement {
        $$ = std::make_shared<IfThenNode>($3, $5, $7,
                                     $1.first, $1.second);
    }
	| SWITCH '(' expression ')' statement {
        $$ = std::make_shared<SwitchNode>($3, $5, $1.first, $1.second);
    }
	;

iteration_statement
	: WHILE '(' expression ')' statement {
        $$ = std::make_shared<WhileNode>($3, $5, false, $1.first, $1.second);
    }
	| DO statement WHILE '(' expression ')' ';' {
        $$ = std::make_shared<WhileNode>($5, $2, true, $1.first, $1.second);
    }
	| FOR '(' expression_statement expression_statement ')' statement {
        $$ = std::make_shared<ForNode>($3, $4, ExprNode::NULL_VAL, $6,
                                  $1.first, $1.second);
    }
	| FOR '(' expression_statement expression_statement expression ')' statement {
        $$ = std::make_shared<ForNode>($3, $4, $5, $7,
                                  $1.first, $1.second);
    }
	;

jump_statement
	: GOTO IDENTIFIER ';' {
        auto label = std::make_shared<IdentNode>($2.get_value(), $2.get_line(), $2.get_col());
        $$ = std::make_shared<GotoNode>(label, $1.first, $1.second);
    };
	| CONTINUE ';' { $$ = std::make_shared<ContinueNode>($1.first, $1.second); }
	| BREAK ';'    { $$ = std::make_shared<BreakNode>($1.first, $1.second); }
	| RETURN ';'   {
        $$ = std::make_shared<ReturnNode>(ExprNode::NULL_VAL,
                                          $1.first, $1.second); }
	| RETURN expression ';'  {
        $$ = std::make_shared<ReturnNode>($2,
                                          $1.first, $1.second); }
	;

translation_unit
	: external_declaration                    {
        prog.declList.push_back($1);
        prog.linenum = $1->get_line();
        prog.colnum =  $1->get_col();
    }
	| translation_unit external_declaration {
        // Note: first token should already have decls in program
        prog.declList.push_back($2);
    }
	;

external_declaration
	: function_definition { $$ = $1; }
	| declaration         { $$ = $1; }
	;

function_definition
	: declaration_specifiers declarator declaration_list compound_statement {
        // K&R-style (initializer list) function decl;
        $$ = nullptr;
        error("Function definitions using initializer lists (K&R-style) is currently not implemented");
    }
	| declaration_specifiers declarator compound_statement {
        // ANSI-style (parameter type list) function decl
        $$ = std::make_shared<FuncDefNode>($1, $2, $3,
                                            $1.front()->get_line(),
                                            $1.front()->get_col());
    }
	| declarator declaration_list compound_statement {
        // K&R-style (initializer list) function decl
        // with default return value of int
        $$ = nullptr;
        error("Function definitions using initializer lists (K&R-style) is currently not implemented");
    }
	| declarator compound_statement {
        // ANSI-style (parameter type list) function decl
        // with default return value of int
        auto default_type_list = FuncDefNode::DeclSpecifierList();
        auto default_type = std::make_shared<IntTypeNode>($1->get_line(),
                                                          $1->get_col());
        default_type_list.push_back(default_type);
        $$ = std::make_shared<FuncDefNode>(default_type_list, $1, $2,
                                            $2->get_line(),
                                            $2->get_col());
    }
	;

%%
#include <stdio.h>

extern char yytext[];
extern int column;

void Parser::error(const std::string& msg)
{
	std::cout << "Line " << Pos::linenum << ": " << msg << std::endl;
}
