#include <iostream>
#include <fstream>
#include "token/token.hpp"
#include "scanner_shim.hpp"
#include "quant_c_parse.hpp"
#include "util/pos.hpp"

#define MIN_ARGC 2
#define EOF_TOKEN 0

using namespace quant_c_token;
using namespace quant_c_compiler;
using namespace util;

int main(int argc, char *argv[]) {

    if (argc < MIN_ARGC) {
        std::cerr << "Expected one argument; the name of a file to scan" << std::endl;
        return -1;
    }

    // Open the file to scan
    std::cout << "Reading file\n";
    std::ifstream quant_c_file;
    try {
        quant_c_file.open(argv[1]);
    } catch (...) {
        std::cerr << "Unable to open file \"" << argv[1] << "\"\n";
        return -1;
    }

    // Create the scanner and begin outputting
    std::cout << "Creating scanner\n";
    Scanner test_scanner(quant_c_file, std::cout);

    // token_type indicates type value of token (defined by parser)
    // semantic_type is a variant that can contain different kinds of object
    // (in this case, they would be Token objects defined in src/token)
    int token_id = -1;
    Parser::semantic_type semantic_value;

    std::cout << "Scanning file\n";
    while (token_id != EOF_TOKEN) {
        token_id = test_scanner.lex(&semantic_value);

        std::cout << "Line: " << Pos::linenum
                  << " Column: " << Pos::colnum << " ";
        std::cout << get_token_name(token_id);

        // Print off values of tokens with additional semantic values
        // (such as literals)
        switch(token_id) {
            case Parser::token_type::IDENTIFIER:{
                std::cout << " name : \"" <<
                semantic_value.as<IdentifierToken>().get_value()
                << "\"";
            }
            break;
            case Parser::token_type::INT_LITERAL:{
                std::cout << " value : \"" <<
                semantic_value.as<IntLitToken>().get_value()
                << "\"";
            }
            break;
            case Parser::token_type::FLOAT_LITERAL:{
                std::cout << " value : \"" <<
                semantic_value.as<FloatLitToken>().get_value()
                << "\"";
            }
            break;
            case Parser::token_type::BOOL_LITERAL:{
                std::cout << " value : \"" <<
                semantic_value.as<BoolLitToken>().get_value()
                << "\"";
            }
            break;
            case Parser::token_type::CHAR_LITERAL:{
                std::cout << " value : \"" <<
                semantic_value.as<CharLitToken>().printCharValue()
                << "\"";
            }
            break;
            case Parser::token_type::STRING_LITERAL:{
                std::cout << " value : \"" <<
                semantic_value.as<StringLitToken>().get_value()
                << "\"";
            }
            break;
            default: break;
        }


        std::cout << "\n";
    }

    // Close file
    try {
        quant_c_file.close();
    } catch (...) {
        std::cerr << "Error while closing file \"" << argv[1] << "\"\n";
        return -1;
    }

    return 0;
}
