#include <iostream>
#include <fstream>
#include "token/token.hpp"
#include "ast/ast.hpp"
#include "scanner_shim.hpp"
#include "quant_c_parse.hpp"

#define MIN_ARGC 2
#define EOF_TOKEN 0

using namespace quant_c_token;
using namespace quant_c_compiler;
using namespace quant_c_ast;

int main(int argc, char *argv[]) {

    if (argc < MIN_ARGC) {
        std::cerr << "Expected one argument; the name of a file to parse" << std::endl;
        return -1;
    }

    // Open the file to scan
    std::cout << "Reading file\n";
    std::ifstream quant_c_file;
    try {
        quant_c_file.open(argv[1]);
    } catch (...) {
        std::cerr << "Unable to open file \"" << argv[1] << "\"\n";
        return -1;
    }

    // Create the parser and parse the file
    std::cout << "Creating scanner\n";
    Scanner test_scanner(quant_c_file, std::cout);
    std::cout << "Creating parser\n";
    TranslationUnitNode ast;
    Parser test_parser(test_scanner, ast);

    /* Uncomment to turn on extensive parser debugging */
    test_parser.set_debug_stream(std::cout);
    test_parser.set_debug_level(1);

    int parse_result = test_parser.parse();

    if (parse_result == 0) {
        std::cout << "Parsing successfully completed.\n";
        ast.unparse(0);
    }
    else {
        std::cout << "Error code " << parse_result << ": Parsing ended early due to errors.\n";
    }

    // Close file
    try {
        quant_c_file.close();
    } catch (...) {
        std::cerr << "Error while closing file \"" << argv[1] << "\"\n";
        return -1;
    }

    return 0;
}
